/* $Id$ */

/*
 * hed - Hexadecimal editor
 * Copyright (C) 2004  Petr Baudis <pasky@ucw.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * 'You see: Mr. Drogo, he married poor Miss Primula Brandybuck. She was our
 * Mr. Bilbo's first cousin on the mother's side (her mother being the youngest
 * of the Old Took's daughters); and Mr. Drogo was his second cousin. So Mr.
 * Frodo is his first _and_ second cousin, once removed either way, as the
 * saying is, if you follow me.
 */

#include "config.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hed.h"

#include "config/config.h"
#include "eval/eval.h"
#include "util/lists.h"


/* This is really just trivial implementation. */

struct atom {
	LIST_HEAD(struct atom);

	int len; // Spans over n bytes; -1 == auto (defaults to 1 at root)

	enum atom_type {
		ATOM_UN,
		ATOM_BIN,
		ATOM_BYTESTR,
		ATOM_NUMBER,
		ATOM_REGISTER,
		ATOM_T_MAX
	} type;
	void *data;
};

typedef void (*atom_evaluator)(struct atom *atom, unsigned char *scramble);
static void atom_un(struct atom *atom, unsigned char *scramble);
static void atom_bin(struct atom *atom, unsigned char *scramble);
static void atom_bytestr(struct atom *atom, unsigned char *scramble);
static void atom_number(struct atom *atom, unsigned char *scramble);
static void atom_register(struct atom *atom, unsigned char *scramble);
static atom_evaluator evals[ATOM_T_MAX] = {
	atom_un, atom_bin, atom_bytestr, atom_number, atom_register,
};

typedef void (*atom_free)(struct atom *atom);
static void free_un(struct atom *atom);
static void free_bin(struct atom *atom);
static void free_bytestr(struct atom *atom);
static void free_number(struct atom *atom);
static void free_register(struct atom *atom);
static atom_free frees[ATOM_T_MAX] = {
	free_un, free_bin, free_bytestr, free_number, free_register,
};

struct atom_un {
	enum atom_un_op {
		AUO_NEG,
	} op;
	struct atom *atom;
};

struct atom_bin {
	enum atom_bin_op {
		ABO_OR,
		ABO_AND,
		ABO_XOR,
		ABO_PLUS,
		ABO_MINUS,
		ABO_MUL,
		ABO_DIV,
		ABO_MOD,
		ABO_SHL,
		ABO_SHR,
	} op;
	struct atom *atom1;
	struct atom *atom2;
};

struct atom_bytestr {
	unsigned char bytes[0]; /* [len] */
};

struct atom_number {
	ubint num; /* maximal size */
};

struct expression;

struct atom_register {
	char reg;
	struct expression *offset;
};

struct expression {
	struct list_head atoms;
	int len;
};


static void
free_atom(struct atom *atom)
{
	frees[atom->type](atom);
	free(atom);
}


static char
unhex(char x)
{
	unsigned char hx[] = "0123456789abcdef";
	return (unsigned char*) strchr(hx, tolower(x)) - (unsigned char *)&hx;
}

static void
skip_white(char **sexpr)
{
	while (isspace(**sexpr))
		(*sexpr)++;
}

static struct atom *atom_parse(char **sexpr);
static struct expression *expr_compile_till(char **sexpr, char till);

static struct atom *
val_parse(char **sexpr)
{
	struct atom *a = NULL;

	skip_white(sexpr);

	if (**sexpr == '(') {
		(*sexpr)++;
		a = atom_parse(sexpr);

	} else if (**sexpr == '~') {
		struct atom *sa;
		(*sexpr)++;
		sa = val_parse(sexpr);
		if (sa) {
			a = calloc(1, sizeof(struct atom));
			a->len = sa->len;
			a->type = ATOM_UN;
			a->data = calloc(1, sizeof(struct atom_un));
			((struct atom_un *) a->data)->atom = sa;
		}

	} else if (**sexpr == '"') {
		char reg;

		(*sexpr)++;
		reg = **sexpr;
		(*sexpr)++;

		a = calloc(1, sizeof(struct atom));
		a->len = -1;
		a->type = ATOM_REGISTER;
		a->data = calloc(1, sizeof(struct atom_register));
		((struct atom_register *) a->data)->reg = reg;

		if (**sexpr == '[') {
			(*sexpr)++;
			((struct atom_register *) a->data)->offset = expr_compile_till(sexpr, ']');
			(*sexpr)++;
		}

	} else if (**sexpr == '\'') {
		char *base;
		int i = 0;

		(*sexpr)++;
		base = *sexpr;
		while (**sexpr && **sexpr != '\'') {
			if (**sexpr == '\\') {
				(*sexpr)++;
				if (!**sexpr)
					return NULL;
			}
			i++;
			(*sexpr)++;
		}

		a = calloc(1, sizeof(struct atom));
		a->len = i;
		a->type = ATOM_BYTESTR;
		a->data = calloc(a->len, 1);
		for (i = 0; base < *sexpr; i++, base++) {
			if (*base == '\\')
				base++;
			((struct atom_bytestr *) a->data)->bytes[i] = *base;
		}
		(*sexpr)++;

	} else if (**sexpr == 'x') {
		char *base;
		int i;

		(*sexpr)++;
		base = *sexpr;
		while (isxdigit(**sexpr))
			(*sexpr)++;

		a = calloc(1, sizeof(struct atom));
		a->len = (*sexpr - base + 1) / 2;
		a->type = ATOM_BYTESTR;
		a->data = calloc(a->len, 1);
		for (i = 0; base < *sexpr; i++, base++) {
			// l33t
			((struct atom_bytestr *) a->data)->bytes[i / 2] |= unhex(*base) << ((1-i%2)*4);
		}

	} else if (isdigit(**sexpr)) {
		char *base = *sexpr;
		ubint num = strtoull(*sexpr, sexpr, 0);
		assert(sizeof(ubint) == sizeof(unsigned long long));
		// ...or fix me to use something better than strtoull()!
		assert(*sexpr > base);

		a = calloc(1, sizeof(struct atom));
		if (base[0] == '0' && base[1] == 'x') {
			/* Take leading zeros of hex numbers into account */
			a->len = (*sexpr - base - 2 + 1) / 2;
		} else {
			/* Whatever fits */
			a->len = 1 + bitsize(num) / 8;
		}
		a->type = ATOM_NUMBER;
		a->data = calloc(1, sizeof(struct atom_number));
		((struct atom_number *) a->data)->num = num;
	}

	if (**sexpr == '{') {
		int len;
		(*sexpr)++;
		len = strtoull(*sexpr, sexpr, 0);
		if (**sexpr != '}') {
			free_atom(a);
			return NULL;
		}
		if (a->type != ATOM_NUMBER || a->len > sizeof(ubint)) {
			/* Not supported yet */
			free_atom(a);
			return NULL;
		}
		a->len = len;
		(*sexpr)++;
	}
	return a;
}

static struct atom *
atom_lparse(char **sexpr, struct atom *sa1)
{
	struct atom *a, *sa2;
	char bopc[] = "|&^+-*/%<>"; // in atom_bin_op order
	char *bop;

	skip_white(sexpr);
	if (!**sexpr)
		return sa1;
	if (**sexpr == ')') {
		(*sexpr)++;
		return sa1; // *SLAMM*
	}
	bop = strchr(bopc, **sexpr);
	if (!bop)
		return sa1;
	(*sexpr)++;
	if (*bop == '<' || *bop == '>') {
		if (*(*sexpr)++ != *bop)
			return NULL;
	}
	sa2 = val_parse(sexpr);
	if (**sexpr == ')')
		(*sexpr)++;
	if (!sa2)
		//fprintf(stderr, "parse error at <%s>\n", *sexpr);
		return NULL;

	if (sa1->len == -1)
		sa1->len = sa2->len;
	else if (sa2->len == -1)
		sa2->len = sa1->len;
	/* Note that BOTH may be -1, but that's ok */

	a = calloc(1, sizeof(struct atom));
	a->len = max(sa1->len, sa2->len);
	a->type = ATOM_BIN;
	a->data = calloc(1, sizeof(struct atom_bin));
	((struct atom_bin *) a->data)->op = bop - bopc;
	((struct atom_bin *) a->data)->atom1 = sa1;
	((struct atom_bin *) a->data)->atom2 = sa2;
	/* Try to expand the epxression */
	return atom_lparse(sexpr, a) ? : a;
}

static struct atom *
atom_parse(char **sexpr)
{
	struct atom *a, *sa1;

	sa1 = val_parse(sexpr);
	if (!sa1) {
		//fprintf(stderr, "parse error at <%s>\n", *sexpr);
		(*sexpr)++; // no luck, maybe next char
		return NULL;
	}
	a = atom_lparse(sexpr, sa1);
	if (!a) {
		free_atom(sa1);
		return NULL;
	}
	return a;
}

static struct expression *
expr_compile_till(char **sexpr, char till)
{
	struct expression *expr = calloc(1, sizeof(struct expression));
	init_list(expr->atoms);
	while (**sexpr != till) {
		struct atom *a = atom_parse(sexpr);
		if (a) {
			add_to_list_end(expr->atoms, a);
			if (a->len == -1)
				a->len = 1;
			expr->len += a->len;
		}
	}
	if (list_empty(expr->atoms)) {
		free(expr);
		expr = NULL;
	}
	return expr;
}

struct expression *
expr_compile(char *sexpr)
{
	return expr_compile_till(&sexpr, '\0');
}


int
expr_len(struct expression *expr)
{
	assert(expr && expr->len >= 0);
	return expr->len;
}


/* FIXME: These two assume LE bytesex */

ubint
bytestr2num(unsigned char *bytestr, int len)
{
	ubint num = 0;
	int i;
	assert(len <= sizeof(ubint));
	for (i = 0; i < len; i++)
		num |= bytestr[i] << ((len - i - 1) * 8);
	return num;
}

void
num2bytestr(unsigned char *bytestr, int len, ubint num)
{
	assert(len <= sizeof(ubint));
	while (len--)
		*(bytestr++) = (num >> (len * 8)) & 0xff;
}


static void
atom_eval(struct atom *parent, struct atom *atom, unsigned char *scramble)
{
	if (parent && atom->len == -1)
		atom->len = parent->len;
	evals[atom->type](atom, scramble);
}

static void
atom_un(struct atom *atom, unsigned char *scramble)
{
	struct atom_un *data = atom->data;
	int i;
	assert(data);
	assert(data->op == AUO_NEG);
	assert(data->atom);
	atom_eval(atom, data->atom, scramble);
	for (i = 0; i < atom->len; i++)
		scramble[i] = ~scramble[i];
}

static void
atom_bin(struct atom *atom, unsigned char *scramble)
{
	struct atom_bin *data = atom->data;
	unsigned char *s1, *s2;
	int i; ubint n1, n2;
	assert(data);
	assert(data->atom1 && data->atom2);
	assert(data->atom1->len <= atom->len && data->atom2->len <= atom->len);
	s1 = calloc(1, atom->len);
	s2 = calloc(1, atom->len);
	atom_eval(atom, data->atom1, s1);
	atom_eval(atom, data->atom2, s2);
	switch (data->op) {
		case ABO_OR:
			for (i = 0; i < atom->len; i++) {
				scramble[i] = s1[i] | s2[i];
			}
			break;
		case ABO_AND:
			for (i = 0; i < atom->len; i++)
				scramble[i] = s1[i] & s2[i];
			break;
		case ABO_XOR:
			for (i = 0; i < atom->len; i++)
				scramble[i] = s1[i] ^ s2[i];
			break;
		case ABO_PLUS:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 + n2);
			break;
		case ABO_MINUS:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 - n2);
			break;
		case ABO_MUL:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 * n2);
			break;
		case ABO_DIV:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 / n2);
			break;
		case ABO_MOD:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 % n2);
			break;
		case ABO_SHL:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 << n2);
			break;
		case ABO_SHR:
			n1 = bytestr2num(s1, data->atom1->len);
			n2 = bytestr2num(s2, data->atom2->len);
			num2bytestr(scramble, atom->len, n1 >> n2);
			break;
		default:
			assert(0);
			break;
	}
}

static void
atom_bytestr(struct atom *atom, unsigned char *scramble)
{
	struct atom_bytestr *data = atom->data;
	assert(data);
	memcpy(scramble, data->bytes, atom->len);
}

static void
atom_number(struct atom *atom, unsigned char *scramble)
{
	struct atom_number *data = atom->data;
	assert(data);
	assert(atom->len <= sizeof(data->num));
	num2bytestr(scramble, atom->len, data->num);
}

static reg_callback reg_cb;
static void *reg_cb_data;
static int reg_cb_pos;

static void
atom_register(struct atom *atom, unsigned char *scramble)
{
	struct atom_register *data = atom->data;
	assert(data);
	if (reg_cb) {
		off_t ofs = 0;
		if (data->offset) {
			unsigned char *ofs_buf = expr_eval(data->offset, reg_cb, reg_cb_data);
			if (ofs_buf) {
				ofs = bytestr2num(ofs_buf, expr_len(data->offset));
				free(ofs_buf);
			}
		}
		reg_cb(reg_cb_data, data->reg, ofs, reg_cb_pos, scramble, atom->len);
	} else {
		memset(scramble, 0, atom->len);
	}
}


unsigned char *
expr_eval(struct expression *expr, reg_callback cb, void *data)
{
	unsigned char *buf;
	struct atom *a;
	int pos = 0;

	reg_cb = cb;
	reg_cb_data = data;

	assert(expr && expr->len >= 0);
	buf = malloc(expr->len);
	foreach (a, expr->atoms) {
		reg_cb_pos = pos;
		atom_eval(NULL, a, buf + pos);
		pos += a->len;
	}
	assert(pos == expr->len);
	return buf;
}


static void
free_un(struct atom *atom)
{
	free(atom->data);
}

static void
free_bin(struct atom *atom)
{
	struct atom_bin *data = atom->data;
	free(data->atom1);
	free(data->atom2);
	free(atom->data);
}

static void
free_bytestr(struct atom *atom)
{
	free(atom->data);
}

static void
free_number(struct atom *atom)
{
	free(atom->data);
}

static void
free_register(struct atom *atom)
{
	free(atom->data);
}
