/* $Id$ */

/*
 * hed - Hexadecimal editor
 * Copyright (C) 2004  Petr Baudis <pasky@ucw.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * There hammer on the anvil smote,
 * There chisel clove, and graver wrote;
 * There forged was blade, and bound was hilt;
 * The delver mined, the mason built.
 * There beryl, pearl, and opal pale,
 * And metal wrought like fishes' mail,
 * Buckler and corslet, axe and sword,
 * And shining spears were laid in hoard.
 */

#include "config.h"

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "hed.h"

#include "file/file.h"

/*
 * `Piles of jewels?' said Gandalf. `No. The Orcs have often plundered Moria;
 * there is nothing left in the upper halls. And since the dwarves fled, no one
 * dares to seek the shafts and treasuries down in the deep places: they are
 * drowned in water--or in a shadow of fear.'
 */


#define BLKGETSIZE 0x1260
#define BLKGETSIZE64 0x1272


/* TODO: Currently the file blocks allocation is not very sophisticated and
 * when the weather is bad it could probably have rather horrible results. */

//#define BLOCKS_DEBUG
#ifdef BLOCKS_DEBUG
#define BDEBUG(x...) fprintf(stderr, x)
#else
#define BDEBUG(x...)
#endif


static char *
swp_filename(const unsigned char *filename)
{
	static char swp[PATH_MAX];
	char *file;

	if (strlen(filename) > PATH_MAX - 8)
		return NULL;
	strcpy(swp, filename);

	file = strrchr(swp, '/');
	file = file ? file + 1 : swp;
	memmove(file + 1, file, strlen(file) + 1);
	*file = '.';
	strcat(file, ".hedswp");
	return swp;
}

static struct file_block *
new_virt_block(off_t size, bool physical)
{
	struct file_block *new = calloc(1, sizeof(struct file_block));
	new->virtual = true;
	new->excache = true;
	new->t.size = size;
	if (physical) new->t.phys_size = size;
	BDEBUG("Spawned new virtual block [%llx] (phys?%d)\n", size, physical);
	return new;
}

static bool
kill_block_if_empty(struct file *file, struct file_block *block)
{
	if (block->t.size == 0 && block->t.phys_size == 0) {
		/* No recalculation needed, zero size. */
		unchain_block(file, block);
		assert(block->excache);
		free(block);
		return true;
	}
	return false;
}

static struct file *
file_basic_init(const unsigned char *name)
{
	int fd;
	struct file *file;
	struct stat statbuf;
	off_t size;

	fd = open(name, O_RDONLY);
	if (fd < 0) {
		if (errno != ENOENT)
			return NULL;
		fprintf(stderr, "Warning: File %s does not exist\n", name);
		memset(&statbuf, 0, sizeof(statbuf));

	} else if (fstat(fd, &statbuf) < 0) {
		int errsv = errno;
		close(fd);
		errno = errsv;
		return NULL;

	} else if (S_ISBLK(statbuf.st_mode)) {
		if (ioctl(fd, BLKGETSIZE64, &size)) {
			int size32;
			if (ioctl(fd, BLKGETSIZE, &size32)) {
				int errsv = errno;
				close(fd);
				errno = errsv;
				return NULL;
			}
			size = size32;
		}

	} else if (S_ISREG(statbuf.st_mode)) {
		size = statbuf.st_size;

	} else {
		close(fd);
		errno = EINVAL;
		return NULL;
	}

	file = calloc(1, sizeof(struct file));
	if (!file) return NULL;
	file->name = name;
	file->fd = fd;
	file->s = statbuf;
	file->size = size;
	init_block_list(file);

	return file;
}

struct file *
file_init(const unsigned char *name)
{
	struct file *file;
	char *swp;

	assert(name);

	swp = swp_filename(name);
	if (!access(swp, F_OK)) {
		fprintf(stderr, "Error: Swap file %s for file %s exists\n", swp, name);
		fprintf(stderr, "This means that hed did not quit properly the last time\n");
		fprintf(stderr, "you edited this file. You can use the swap file to restore\n");
		fprintf(stderr, "some of your unsaved work. Run hed with --restore-swp to\n");
		fprintf(stderr, "do that (until you save your file, the swap file will\n");
		fprintf(stderr, "stay around). Run hed with --remove-swp to get rid of\n");
		fprintf(stderr, "the file. (You can also use both options simultanously.)\n");
		errno = 0;
		return NULL;
	}

	file = file_basic_init(name);
	if (file && file_size(file)) {
		struct file_block *block;
		block = new_virt_block(file_size(file), true);
		chain_block(file, block);
	}

	return file;
}

void
dump_block(int level, struct file_block *block, off_t *cur_offset, off_t *cur_poffset)
{
	char t[21] = "                    ";
	if (block->t.left)
		dump_block(level + 1, (struct file_block *) block->t.left, cur_offset, cur_poffset);
	if (level < 20) t[level] = '>'; else t[19] = '.';
	fprintf(stderr, "%s [%06llx] [%06llx] %c %c %05llx %05llx {%02x%02x%02x%02x} -- %p ^%p [%06llx] [%06llx]\n",
	        t, *cur_offset, *cur_poffset,
		block->virtual ? 'v' : ' ', block->dirty ? '*' : ' ',
		block->t.size, block->t.phys_size,
		!block->virtual && block->t.size > 0 ? block->src[0] : 0,
		!block->virtual && block->t.size > 1 ? block->src[1] : 0,
		!block->virtual && block->t.size > 2 ? block->src[2] : 0,
		!block->virtual && block->t.size > 3 ? block->src[3] : 0,
		block, block->t.up,
		block->t.cover_size, block->t.cover_phys_size
		);
	*cur_offset += block->t.size;
	*cur_poffset += block->t.phys_size;
	if (block->t.right)
		dump_block(level + 1, (struct file_block *) block->t.right, cur_offset, cur_poffset);
	assert(block->t.cover_size == (block->t.left ? ((struct file_block*)block->t.left)->t.cover_size : 0)
	                            + (block->t.right ? ((struct file_block*)block->t.right)->t.cover_size : 0)
				    + block->t.size);
	assert(block->t.cover_phys_size == (block->t.left ? ((struct file_block*)block->t.left)->t.cover_phys_size : 0)
	                                 + (block->t.right ? ((struct file_block*)block->t.right)->t.cover_phys_size : 0)
					 + block->t.phys_size);
}

void
dump_blocks(struct file *file)
{
#ifdef BLOCKS_DEBUG
	off_t cur_offset = 0, cur_poffset = 0;

	fprintf(stderr, "-- blocks dump --\n");
	/* foreach_block() splays things around and that is disruptive.
	 * We will do it manually for the trees here: */
	dump_block(0, file->blocks.root, &cur_offset, &cur_poffset);
	fprintf(stderr, "-- blocks dump end --\n");
#endif
}

/* Returns pointer to the _following_ block after this block
 * is evaporated. */
static struct file_block *
file_kill_block(struct file *file, struct file_block *block)
{
	struct file_block *prev = prev_block(file, block);
	struct file_block *next = next_block(file, block);

	/* We should never kill a dirty block! */
	assert(!block->dirty);
	/* The block is not dirty, thus it must not be an
	 * insertion/deletion. */
	assert(block->t.size == block->t.phys_size);
	/* We shouldn't get with an empty block here (that might
	 * need special considerations with virtualization). */
	assert(block->t.size > 0);

	/* Messing with block sizes and unchaining is a bit tricky
	 * since we can splay() as a side effect of some innocent
	 * operation (like next_block()) and propagate the interim
	 * wrong block sizes all around. What is more troublesome,
	 * unchain_block() _itself_ can splay(). So we really need
	 * to recalc_node_recursive() right after we update the size.
	 * If this place turns out to be a hot-spot, we can optimize
	 * the tree operations here. */

	if (is_a_block(file, prev) && prev->virtual) {
		if (is_a_block(file, next) && next->virtual && !next->protect) {
			struct file_block *next2;
			next2 = next_block(file, next);

			assert(next->t.size == next->t.phys_size);
			prev->t.size += next->t.size;
			prev->t.phys_size += next->t.phys_size;
			recalc_node_recursive(prev);
			recalc_unchain_block(file, next);
			assert(next->excache);
			free(next);

			next = next2;
		}
		/* Asserted that won't break the size consistency */
		prev->t.size += block->t.size;
		prev->t.phys_size += block->t.phys_size;
		recalc_node_recursive(prev);

	} else if (is_a_block(file, next) && next->virtual) {
		/* Asserted that won't break the size consistency */
		next->t.size += block->t.size;
		next->t.phys_size += block->t.phys_size;
		recalc_node_recursive(next);

	} else {
		/* Asserted that will not ever kill non-physical blocks */
		struct file_block *gap = new_virt_block(block->t.size, true);
		recalc_chain_block_after(file, block, gap);
	}

	recalc_unchain_block(file, block);

	return next;
}

static struct file_block *
file_alloc_block(struct file *file)
{
	struct file_block *new;

	file->cache_pos %= CACHE_LENGTH;
	do {
		new = &file->cache[file->cache_pos++];
		/* Protecting is rare and individual so we shouldn't get into
		 * an infinite loop here. */
	} while (new->protect);
	assert(!new->excache);
	assert(!new->virtual);

	if (new->dirty) {
		/* We need to keep it. Put it outside of the cache. */
		struct file_block *dup;

		dup = calloc(1, sizeof(struct file_block));
		*dup = *new;
		dup->excache = true;
		new->dirty = false;

		assert(is_chained_block(file, new));
		recalc_chain_block_after(file, new, dup);
		recalc_unchain_block(file, new);

	} else if (is_chained_block(file, new)) {
		/* We will overwrite already used cache block. */
		file_kill_block(file, new);
	}

	return new;
}

static struct file_block *
file_load_block(struct file *file, off_t phys_offset, size_t size)
{
	/* FIXME: We waste CPU by copying the buffer around. */
	unsigned char buf[FILE_BLOCK_SIZE];
	struct file_block *new;
	ssize_t rsize;
	
	assert(size <= FILE_BLOCK_SIZE);

	if (file->fd < 0)
		return NULL;

	rsize = pread(file->fd, buf, size, phys_offset);
	if (rsize <= 0) {
		if (rsize == 0) errno = 0;
		return NULL;
	}

	new = file_alloc_block(file);
	new->t.size = new->t.phys_size = rsize;
	memcpy(new->src, buf, rsize);
	BDEBUG("Loaded block %llx:(%x,%llx)\n", phys_offset, size, new->t.size);
	return new;
}

void *
file_fetch_block(struct file *file, off_t offset, ssize_t *size)
{
	struct file_block *block;
	unsigned char *buf;
	/* Absolute offset of the current block. */
	off_t cur_offset = 0;
	/* cur_offset for phys_offset. */
	off_t cur_poffset = 0;
	off_t buf_offset = 0;
	ssize_t todo_size;

	assert(file);
	assert(offset >= 0 && size && *size >= 0);

	buf = calloc(1, *size);
	todo_size = *size;

	BDEBUG("Fetching block %llx:%x\n", offset, *size);
	dump_blocks(file);

	block = find_in_tree(&file->blocks, offset);
	if (is_a_block(file, block)) {
		cur_offset = tree_block_offset(block);
		cur_poffset = tree_block_phys_offset(block);
		BDEBUG("Selected block %llx:%llx\n", cur_offset, block->t.size);
		dump_blocks(file);
	} else {
		BDEBUG("Block not found\n");
	}

	while (todo_size > 0) {
		off_t block_offset;
		ssize_t copy_size;

		/* Past the end of file, game over. */
		if (!is_a_block(file, block))
			break;

		BDEBUG("[%llx] Considering block (%llx) %p\n", cur_offset, block->t.size, block->t.up);
		assert(cur_offset + block->t.size > offset);

		block_offset = offset - cur_offset;
		if (block->virtual) {
			struct file_block *prev, *newblock;
			ssize_t read_size;

			if (block->t.phys_size == 0) {
				/* We are accessing a virtual block with
				 * zero physical size. That can happen if
				 * it fills a gap to an unsaved dirty block
				 * added past the end of file. */
				/* We will fill the hole in the buffer
				 * with zeroes (by skipping it). */
				if (todo_size < block->t.size) {
					/* We _end_ in the hole. */
					todo_size = 0;
					break;
				}
				buf_offset += block->t.size;
				offset += block->t.size;
				todo_size -= block->t.size;
				goto continue_loop;
			}

			assert(block->t.phys_size == block->t.size);
			read_size = min(FILE_BLOCK_SIZE, block->t.phys_size - block_offset);
			BDEBUG("****l**** [block %llx %llx/%llx <%p>] %llx->%llx:%x\n",
				cur_poffset, block->t.size, block->t.phys_size, block->t.up,
				offset, cur_poffset + block_offset, read_size);
			assert(read_size > 0);
			BDEBUG("=bs= %llx %p\n", block->t.size, block->t.up);
			/* Without block->protect, we can get to a truly
			 * beautiful situation like this:
			 * 	[virt] something
			 * 	[phys] something
			 * 	[virt] block
			 * file_load_block() will allocate a block which will
			 * kick the [phys] block from cache and that will
			 * subsequently cause the [virt] blocks to get merged
			 * and we will lose out @block. *KABOOM* */
			block->protect++;
			newblock = file_load_block(file, cur_poffset + block_offset, read_size);
			block->protect--;
			if (!newblock) {
				if (errno < 0) {
					int errsv = errno; free(buf); errno = errsv;
					return NULL;
				}

				/* Reading past EOF while phys_size was non-zero... uh-oh? */
				assert(0);
			}
			BDEBUG("=bs= %llx %p\n", block->t.size, block->t.up);

			/* At this moment, the worst-case scenario is:
			 * .........XXXXXXXXXXXX.........
			 *  virtual  we've just  virtual
			 *           read this
			 */

			assert(offset >= cur_offset);
			BDEBUG("=bs= %llx\n", block->t.size);
			prev = prev_block(file, block);
			if (block_offset) {
				/* We have left some leading part of the
				 * block virtual - recreate it as a separate
				 * block. */
				struct file_block *lead;
				lead = new_virt_block(block_offset, true);

				block->t.size -= block_offset;
				block->t.phys_size -= block_offset;
				recalc_node_recursive(block);

				recalc_chain_block_after(file, prev, lead);
				prev = lead;
			}

			/* Chain our loaded block */
			recalc_chain_block_after(file, prev, newblock);

			/* Perhaps we have some tail still-virtual area...
			 * if so, we will reuse our current @block for that. */
			assert(block->t.size >= newblock->t.size);
			block->t.size -= newblock->t.size;
			block->t.phys_size -= newblock->t.phys_size;
			assert(block->t.size >= 0 && block->t.phys_size >= 0);
			assert(block->t.phys_size == block->t.size);
			recalc_node_recursive(block);
			kill_block_if_empty(file, block);

			dump_blocks(file);

			/* Make us insert the newly loaded block to the buffer now. */
			block = newblock;
			cur_offset += block_offset;
			cur_poffset += block_offset;
		}
		assert(block->t.size > 0);

		block_offset = offset - cur_offset;
		BDEBUG("src %llx = %llx - %llx:%llx\n", block_offset, offset, cur_offset, block->t.size);
		assert(block_offset < block->t.size && block_offset >= 0);
		copy_size = min(todo_size, block->t.size - block_offset);
		assert(copy_size > 0);
		memcpy(buf + buf_offset, block->src + block_offset, copy_size);

		buf_offset += copy_size;
		offset += copy_size;
		todo_size -= copy_size;

		/* Everybody loves Perl and its continue block. */
continue_loop:
		assert(is_a_block(file, block));
		cur_offset += block->t.size;
		cur_poffset += block->t.phys_size;
		block = next_block(file, block);
	}
	*size -= todo_size;

	return buf;
}


struct ffb_hookdata {
	struct file *file;
	off_t pos;
	unsigned char *buf;
	int blen;
	reg_callback base_ecb;
	void *base_ecb_data;
};

static void
eval_reg_cb(void *hookdata, char reg, bint ofs, int mpos, unsigned char *scramble, int len)
{
	struct ffb_hookdata *data = hookdata;
	if (reg == '.') {
		if (ofs >= 0 && (ofs + len <= data->blen - mpos)) {
			memcpy(scramble, data->buf + ofs + mpos, len - ofs);
		} else {
			/* Oh the horrors, we look _outside_ of the match now. */
			/* FIXME: Optimize */
			int blen = len;
			unsigned char *buf = file_fetch_block(data->file, max(0, data->pos), &blen);
			memcpy(scramble, buf, blen);
			memset(scramble, 0, len - blen);
			free(buf);
		}
	} else {
		data->base_ecb(data->base_ecb_data, reg, ofs, mpos, scramble, len);
	}
}

off_t
file_find_expr(struct file *file, off_t from,
		struct expression *expr, reg_callback expr_cb, void *expr_cb_data,
		int dir)
{
	/* TODO: Use custom iteration routine instead fetching blocks for each
	 * match. Should be much faster. */
	struct ffb_hookdata data;
	int len = expr_len(expr);

	assert(dir == 1 || dir == -1);
	assert(len > 0);

	data.file = file;
	data.base_ecb = expr_cb;
	data.base_ecb_data = expr_cb_data;

	while (0 <= from && from < file->size - len) {
		unsigned char *seq;
		int pos;

		data.blen = len;
		data.buf = file_fetch_block(file, from, &data.blen);
		if (data.blen != len)
			return FINDOFF_NO_MATCH;
		data.pos = from;
		seq = expr_eval(expr, eval_reg_cb, &data);
		for (pos = 0; pos < len; pos++)
			if (seq[pos] != data.buf[pos])
				break;
		free(seq);
		if (pos == len)
			return from;
		from += dir;
	}

	return FINDOFF_NO_MATCH;
}


static struct file_block *
file_get_block_in_virtual(struct file *file,
                          struct file_block *block, off_t cur_offset,
                          off_t block_offset, off_t *abs_offset, bool insert)
{
	bool append_to_prev = false;
	struct file_block *prev = prev_block(file, block);
	if (is_a_block(file, prev))
		BDEBUG("[prev] virt %d size %llx/%llx\n", prev->virtual, prev->t.size, prev->t.phys_size);
	else
		BDEBUG("[prev] not a block\n");

	if (is_a_block(file, prev) && prev->t.size + block_offset < FILE_BLOCK_SIZE
	    && prev->dirty && !!prev->t.phys_size == !!block->t.phys_size) {
		/* We can chop the virtual segment before our point and
		 * accomodate it in the previous block. */
		BDEBUG("Appending to previous block\n");
		assert(!prev->virtual);
		append_to_prev = true;
	}

	/* TODO: Check whether we can append to the next block; otherwise
	 * a pathological case of many single-byte blocks might happen.
	 * --pasky */

	if (append_to_prev || !block_offset) {
		/* Shorten our current virtual block by the
		 * leading block up to block_offset. */
		block->t.size -= block_offset + !insert;
		if (block->t.phys_size > 0)
			block->t.phys_size -= block_offset + !insert;
		recalc_node_recursive(block);
		kill_block_if_empty(file, block);
		/* We will be appending to/after previous block. */

	} else {
		/* We are in the middle of a virtual and cannot
		 * append leading part to the previous block.
		 * Split the virtual block so that we can create
		 * a tiny non-virtual block in the middle for us. */
		if (block->t.size - block_offset - !insert > 0) {
			struct file_block *new;
			new = new_virt_block(block->t.size - block_offset - !insert, !!block->t.phys_size);
			recalc_chain_block_after(file, block, new);
		}

		block->t.size = block_offset;
		if (block->t.phys_size > 0)
			block->t.phys_size = block_offset;
		recalc_node_recursive(block);

		/* We will be appending after @block. */
		prev = block;
	}

	if (append_to_prev) {
		/* Append to the previous block */
		block = prev, cur_offset -= block->t.size;
		block_offset += block->t.size;

		/* FIXME: This is _BROKEN_ and "works" only by accident.
		 * We need to fetch the missing stuff from the file
		 * instead of just zeroing. --pasky */
		memset(block->src + block->t.size, '\0', block_offset - block->t.size + 1);
		block->t.size = block_offset + 1;
		if (block->t.phys_size > 0)
			block->t.phys_size = block_offset + 1;
		recalc_node_recursive(block);

	} else {
		/* Create new block */
		prev->protect++;
		block = file_alloc_block(file);
		prev->protect--;
		block->dirty = true;
		block->t.size = 1;
		/* Punching hole in a gap does not qualify for phys_size. */
		block->t.phys_size = !insert && (is_a_block(file, prev) ? !!prev->t.phys_size : false);
		block->src[0] = '\0';
		if (is_a_block(file, prev))
			recalc_chain_block_after(file, prev, block);
		else
			recalc_chain_block(file, block);
		cur_offset += block_offset;
	}

	dump_blocks(file);
	*abs_offset = cur_offset;
	return block;
}

static struct file_block *
file_get_block(struct file *file, off_t offset, off_t *abs_offset, bool insert)
{
	struct file_block *block;
	/* Absolute offset of the current block. */
	off_t cur_offset = 0;
	/* Relative offset in the current block. */
	off_t block_offset = 0;
	bool trailing_virtual;

	if (insert && file->size < OFF_MAX)
		file->size++;
	if (offset >= file->size)
		file->size = offset + 1;

	BDEBUG("Asking for block at %llx\n", offset);

	block = find_in_tree(&file->blocks, offset);
	if (!is_a_block(file, block)) {
		/* No accomodation found, access after end of file. */
		/* We will simply create a virtual block extending to
		 * just @offset that we will possibly kill a moment
		 * later. */
		struct file_block *last = last_block(file);
		if (is_a_block(file, last))
			cur_offset = tree_block_offset(last) + last->t.size;
		/* else empty file */

		if (is_a_block(file, last) && last->virtual) {
			/* Obviously, if the last block is virtual, we can
			 * just enlarge it. */
			last->t.size += offset - cur_offset + 1;
			block = last;
			recalc_node_recursive(block);
		} else {
			block = new_virt_block(offset - cur_offset + 1, false);
			if (is_a_block(file, last))
				recalc_chain_block_after(file, last, block);
			else
				recalc_chain_block(file, block);
		}

		trailing_virtual = true;
		BDEBUG("Accomodation after EOF, dump:\n");
		dump_blocks(file);
	}

	cur_offset = tree_block_offset(block);
	block_offset = offset - cur_offset;
	BDEBUG("[%llx] block for %llx at %llx, virt %d size %llx/%llx\n",
	       cur_offset, offset, block_offset, block->virtual,
	       block->t.size, block->t.phys_size);
	assert(cur_offset <= offset && cur_offset + block->t.size > offset);
	assert(block_offset >= 0 && block_offset < block->t.size);

	if (block->virtual) {
		struct file_block *new;
		/* We've fallen into a virtual block we will have to "devirtualize". */
		new = file_get_block_in_virtual(file, block, cur_offset, block_offset, abs_offset, insert);

		/* If we are inserting, we did not get rid of the trailing virtual. */
		/* Manual deletion. */
		if (trailing_virtual && insert) {
			recalc_unchain_block(file, block);
			free(block);
		}
		return new;
	}

	/* We can accomodate the byte in our chosen block. */
	assert(block->t.size > block_offset);

	if (insert) {
		/* Insert a byte to the block. */
		if (block->t.size >= FILE_BLOCK_SIZE) {
			struct file_block *next = next_block(file, block);

			/* Can we shift one byte to the next block? */
			if (is_a_block(file, next) && !next->virtual && next->t.size < FILE_BLOCK_SIZE) {
				memmove(next->src, next->src + 1, next->t.size);
				next->src[0] = block->src[FILE_BLOCK_SIZE - 1];
				next->t.size++;
				next->dirty = true;
				recalc_node_recursive(next);
			} else {
				struct file_block *new;
				block->protect++;
				new = file_alloc_block(file);
				block->protect--;
				new->dirty = true;
				new->t.size = 1;
				new->t.phys_size = 0;
				new->virtual = false;
				new->src[0] = block->src[FILE_BLOCK_SIZE - 1];
				recalc_chain_block_after(file, block, new);
				next = new;
			}
			block->t.size--;
			assert(block->t.size < FILE_BLOCK_SIZE);
			if (block->t.phys_size) {
				block->t.phys_size--;
				next->t.phys_size++;
				recalc_node_recursive(next);
			}
			recalc_node_recursive(block);
		}

		memmove(block->src + block_offset + 1, block->src + block_offset, block->t.size - block_offset);
		block->src[block_offset] = '\0';
		block->t.size++;
		block->dirty = true;
		recalc_node_recursive(block);
	}

	dump_blocks(file);
	*abs_offset = cur_offset;
	return block;
}

void
file_set_byte(struct file *file, off_t offset, unsigned char byte)
{
	struct file_block *block;
	off_t abs_offset = 0;

	assert(file && offset >= 0);

	block = file_get_block(file, offset, &abs_offset, false);
	assert(block->t.size > offset - abs_offset);
	block->src[offset - abs_offset] = byte;
	block->dirty = true;
	file->modified = true;
}

void
file_insert_byte(struct file *file, off_t offset, unsigned char byte)
{
	struct file_block *block;
	off_t abs_offset = 0;
	int mark;

	assert(file && offset >= 0);

	block = file_get_block(file, offset, &abs_offset, true);
	assert(block->t.size > offset - abs_offset);
	block->src[offset - abs_offset] = byte;
	block->dirty = true;
	file->modified = true;

	/* Slide the marks */
	for (mark = 0; mark < sizeof(file->marks)/sizeof(file->marks[0]); mark++)
		if (file->marks_set[mark] && file->marks[mark] >= offset)
			file->marks[mark]++;
}

/* FIXME: This is horribly ineffective */
void
file_insert_block(struct file *file, off_t offset, unsigned char *buf, off_t len)
{
	off_t i;
	for (i = 0; i < len; i++)
		file_insert_byte(file, offset + i, buf[i]);
}

void
file_erase_byte(struct file *file, off_t offset)
{
	struct file_block *block;
	off_t block_offset;
	int mark;

	assert(file && offset >= 0);
	if (offset >= file->size)
		return;

	block = find_in_tree(&file->blocks, offset);
	assert(tree_block_offset(block) <= offset);
	block_offset = offset - tree_block_offset(block);

	if (!block->virtual) {
		memmove(block->src + block_offset, block->src + block_offset + 1, block->t.size - block_offset - 1);
		block->dirty = true;
	}
	block->t.size--;
	recalc_node_recursive(block);
	kill_block_if_empty(file, block);
	file->size--;
	file->modified = true;

	/* Slide the marks */
	for (mark = 0; mark < sizeof(file->marks)/sizeof(file->marks[0]); mark++)
		if (file->marks_set[mark]) {
			if (file->marks[mark] > offset)
				file->marks[mark]--;
			else if (file->marks[mark] == offset)
				file->marks_set[mark] = false;
		}
}

/* FIXME: This is horribly ineffective */
void
file_erase_block(struct file *file, off_t offset, off_t len)
{
	off_t i;
	for (i = 0; i < len; i++)
		file_erase_byte(file, offset);
}

void
file_commit(struct file *file)
{
	unsigned char *savebuf = NULL;
	ssize_t overlap = 0;
	ssize_t alloclen = 0;
#define enlarge_savebuf(l_) \
	do { \
		if (l_ > alloclen) { \
			if (alloclen < 1) alloclen = 1; \
			while (l_ > alloclen) alloclen <<= 1; \
			savebuf = realloc(savebuf, alloclen); \
		} \
	} while (0)

	struct file_block *block, *nblock;
	int wfd;
	off_t cur_offset = 0;
	bool clean = true;

	wfd = open(file->name, O_RDWR | (file->fd < 0 ? O_CREAT : 0), 0666);
	if (wfd < 0) {
		perror(file->name);
		return;
	}

	BDEBUG("Going to commit...\n");
	dump_blocks(file);

	/* When we inserted something, we must take care not to overwrite
	 * data ahead of us:
	 *
	 *     X
	 * ---------
	 * ||||\\\\\\
	 * |||| \\\\\\
	 * ----++-----
	 *     XXX
	 *
	 * Until XXX, @overlap is zero, at that moment it is increased to
	 * 2 since that's size-phys_size of the XXX block. But before
	 * committing the XXX block, we must read @overlap bytes BEFORE THE
	 * BLOCK END to @savebuf reuse them later when we might hit a virtual
	 * block. Since now on, we must repeat this with each block (unless
	 * we accumulate enough erases to compensate).
	 *
	 * ---------
	 * |||| ////
	 * ||||////
	 * --------
	 *
	 * When overlap is negative, we merely need to remember to shift
	 * non-dirty blocks along. */
	foreach_block (block, file) {
		off_t todo, n = 0;

#define paranoid_pwrite(fd_, buf_, len_, ofs_) \
	if (pwrite(fd_, buf_, len_, ofs_) < len_) { \
		perror(file->name); clean = false; break; }

		BDEBUG("Committing [%llx %llx/%llx] (dirty?%d)... overlap %llx.\n",
				cur_offset, block->t.size, block->t.phys_size,
				block->dirty, overlap);

		if (overlap == 0 && !block->dirty) {
			/* Skippy */
			assert((block->t.size == block->t.phys_size) || (block->virtual && !block->t.phys_size));

		} else if (!block->dirty) {
			off_t block_offset = 0;
			/* We just carry around some overlap over a
			 * non-dirty block. */
			assert(overlap != 0);
			/* The block itself is totally uninteresting and
			 * might be even virtual (likely is). We care only
			 * about its size. */
			assert((block->t.size == block->t.phys_size) || (block->virtual && !block->t.phys_size));

			if (overlap > 0) {
				/* We will move the data ahead by @overlap, updating
				 * @savebuf on the way. */
				while (block_offset < block->t.size) {
					unsigned char buf[FILE_BLOCK_SIZE];
					ssize_t buflen;

					todo = min(block->t.size - block_offset, FILE_BLOCK_SIZE);
					buflen = pread(wfd, buf, todo, cur_offset + block_offset);
					if (buflen < todo) {
						memset(buf + buflen, '\0', todo - buflen);
						buflen = todo;
					}

					todo = min(overlap, buflen);
					paranoid_pwrite(wfd, savebuf, todo, cur_offset + block_offset);
					if (todo == overlap) {
						paranoid_pwrite(wfd, buf, buflen - todo, cur_offset + block_offset + todo);
						memcpy(savebuf, buf + buflen - todo, todo /* == overlap */);
					} else {
						memmove(savebuf, savebuf + buflen, overlap - buflen);
						memcpy(savebuf + overlap - buflen, buf, buflen);
					}

					block_offset += buflen;
				}

			} else { assert(overlap < 0);
				/* Read data ahead by @overlap and write them back.
				 * @savebuf is irrelevant. */
				while (block_offset < block->t.size) {
					unsigned char buf[FILE_BLOCK_SIZE];
					ssize_t buflen;

					todo = min(block->t.size - block_offset, FILE_BLOCK_SIZE);
					buflen = pread(wfd, buf, todo, cur_offset + block_offset - overlap);
					paranoid_pwrite(wfd, buf, buflen, cur_offset + block_offset);

					block_offset += buflen;
				}
			}

		} else {
			ssize_t odelta;
			assert(!block->virtual);

			odelta = block->t.size - block->t.phys_size;
			if (odelta > 0) {
				overlap += odelta;
				if (overlap > 0)
					enlarge_savebuf(overlap);
			} else if (odelta < 0) {
				overlap += odelta;
				if (overlap > 0)
					memmove(savebuf, savebuf - odelta, overlap);
			}

			if (overlap > 0) {
				/* We have a dirty block to write, which means that
				 * we do not need to use current @savebuf, but we still
				 * do need to back up whatever will we overwrite. */
				todo = min(overlap, block->t.size);
				if (todo < overlap)
					memmove(savebuf, savebuf + todo, overlap - todo);
				assert(overlap >= todo);
				n = pread(wfd, savebuf + overlap - todo, todo, cur_offset + block->t.size - todo);
				if (n < todo)
					memset(savebuf + overlap - todo + n, '\0', todo - n);
			}

			paranoid_pwrite(wfd, block->src, block->t.size, cur_offset);

			/* block->dirty = false; */ /* We will collect garbage later */
		}

		block->t.phys_size = block->t.size;
		recalc_node_recursive(block);
		cur_offset += block->t.size;

#undef paranoid_pwrite
	}

	if (clean) {
		BDEBUG("Truncating at %llx...\n", cur_offset);
		ftruncate(wfd, cur_offset);
		file->modified = false;
	}
	close(wfd);
	if (savebuf)
		free(savebuf);

	file_rmswp(file->name);

	if (file->fd < 0)
		file->fd = open(file->name, O_RDONLY);
	if (file->fd < 0) {
		fprintf(stderr, "Error: File %s cannot be reopened readonly\n", file->name);
		/* This is actually fatal, but we can at least pretend... */
		clean = false; /* Keep the blocks since we won't be able to reload them. */
	}

	/* We need to have this separate, since file_kill_block() can span
	 * following blocks etc. */
	foreachsafe_block (block, nblock, file) {
		if (block->excache && (clean || !block->dirty)) {
			block->dirty = false;
			/* We could have invalidated even nblock.
			 * This might make us go one block back but
			 * that should not matter. */
			nblock = file_kill_block(file, block);
			free(block);
			continue;
		}
		if (clean)
			block->dirty = false;
		kill_block_if_empty(file, block);
	}

#undef enlarge_savebuf
}

static void
file_basic_done(struct file *file)
{
	assert(file);
	if (file->fd >= 0)
		close(file->fd);
	free(file);
}

void
file_done(struct file *file)
{
	file_basic_done(file);
}

static void
write_backlog(off_t backlog, off_t pbacklog, FILE *swp)
{
	char payload_virtual = 0;
	if (!backlog && !pbacklog)
		return;
	fwrite(&payload_virtual, 1, 1, swp);
	fwrite(&backlog, sizeof(backlog), 1, swp);
	fwrite(&pbacklog, sizeof(pbacklog), 1, swp);
}

void
file_swp(struct file *file)
{
	char *swpfile = swp_filename(file->name);
	FILE *swp;
	struct stat s;
	struct file_block *block;
	off_t backlog = 0, pbacklog = 0;
	char ver = 0;
	size_t offsize = sizeof(off_t);
	char payload_dirty = 1;

	if (!swpfile)
		return;
	if (fstat(file->fd, &s))
		return;
	swp = fopen(swpfile, "w");
	if (!swp)
		return;

	fwrite(&ver, 1, 1, swp);
	fwrite(&offsize, sizeof(offsize), 1, swp);
	fwrite(&s, sizeof(s), 1, swp);
	fwrite(&file->size, sizeof(file->size), 1, swp);
	fwrite(&file->modified, sizeof(file->modified), 1, swp);
	fwrite(&file->marks_set, sizeof(file->marks_set), 1, swp);
	fwrite(&file->marks, sizeof(file->marks), 1, swp);

	/* We will join virtual and cached blocks */
	foreach_block (block, file) {
		if (!block->dirty) {
			backlog += block->t.size;
			pbacklog += block->t.phys_size;
			continue;
		}
		assert(!block->virtual);
		write_backlog(backlog, pbacklog, swp);
		fwrite(&payload_dirty, 1, 1, swp);
		fwrite(&block->t.size, sizeof(block->t.size), 1, swp);
		fwrite(&block->t.phys_size, sizeof(block->t.phys_size), 1, swp);
		fwrite(&block->src, block->t.size, 1, swp);
	}
	write_backlog(backlog, pbacklog, swp);

	fclose(swp);
}

struct file *
file_unswp(const unsigned char *name)
{
	char *swpfile = swp_filename(name);
	struct file *file;
	FILE *swp;
	struct stat s;
	char payload;
	size_t offsize;

	errno = 0;
	if (!swpfile) {
		errno = ENAMETOOLONG;
		goto fail_total;
	}

	file = file_basic_init(name);
	if (!file) {
		fprintf(stderr, "Couldn't init file struct\n");
		goto fail_total;
	}

	if (fstat(file->fd, &s)) {
		fprintf(stderr, "stat error\n");
		goto fail_done;
	}
	swp = fopen(swpfile, "r");
	if (!swp) {
		fprintf(stderr, "fopen error\n");
		goto fail_done;
	}

	fread(&payload, 1, 1, swp);
	if (payload != 0) {
		fprintf(stderr, "version mismatch\n");
		goto fail_close;
	}

	fread(&offsize, sizeof(offsize), 1, swp);
	if (offsize != sizeof(off_t)) {
		fprintf(stderr, "off_t size mismatch\n");
		goto fail_close;
	}

	fread(&s, sizeof(s), 1, swp);
	if (memcmp(&s, &file->s, sizeof(s))) {
		fprintf(stderr, "stat info mismatch (you modified the file since hed ran on it; refusing to touch it)\n");
		goto fail_close;
	}

	fread(&file->size, sizeof(file->size), 1, swp);
	fread(&file->modified, sizeof(file->modified), 1, swp);
	fread(&file->marks_set, sizeof(file->marks_set), 1, swp);
	fread(&file->marks, sizeof(file->marks), 1, swp);

	while ((payload = fgetc(swp)) != EOF) {
		struct file_block *block = calloc(1, sizeof(struct file_block));

		if (payload == 0) { // virtual
			block->dirty = false;
			block->excache = true;
			block->virtual = true;
			fread(&block->t.size, sizeof(block->t.size), 1, swp);
			fread(&block->t.phys_size, sizeof(block->t.phys_size), 1, swp);
		} else if (payload == 1) { // physical
			block->dirty = true;
			block->excache = true;
			block->virtual = false;
			fread(&block->t.size, sizeof(block->t.size), 1, swp);
			fread(&block->t.phys_size, sizeof(block->t.phys_size), 1, swp);
			fread(&block->src, block->t.size, 1, swp);
		} else {
			fprintf(stderr, "format error - payload 0x%x\n", payload);
			goto fail_close;
		}
		recalc_chain_block(file, block);
	}

	fclose(swp);
	return file;

fail_close:
	fclose(swp);
	errno = 0;
fail_done:
	{
		int errno_save = errno;
		file_basic_done(file);
		errno = errno_save;
	}
fail_total:
	return NULL;
}

void
file_rmswp(const unsigned char *name)
{
	char *swpfile = swp_filename(name);
	remove(swpfile);
}
