/* The file handling */
/* $Id$ */

/*
 * hed - Hexadecimal editor
 * Copyright (C) 2004  Petr Baudis <pasky@ucw.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef HED__FILE_FILE_H
#define HED__FILE_FILE_H

#include <sys/stat.h>

#include "eval/eval.h"


/* File object */

#define FILE_BLOCK_SIZE 1024
#define CACHE_LENGTH 64
/* Undefine USE_TREE to use the old linear implementation. */
#define USE_TREE

#ifdef USE_TREE
#include "util/tree.h"
#else
#include "util/lists.h"
#endif

/* The block cache is organized as follows:
 *
 * At the start, there is a single virtual block:
 * -
 *
 * Then we can do a read and turn the virtual block to two, or shift its
 * start/end:
 * -x-
 *
 * After a while...
 * x-xxx-xx-x-xxxxxxx-x
 *
 * As you can see, virtual blocks can have any size and stretch variably
 * and fill the gaps between real blocks which have a size limit.
 *
 * Now, we can dirty some blocks:
 * x-xxX-xx-x-xxxxxxx-x
 *
 * Even if we insert anything to the block, other blocks aren't touched,
 * except when the block size hits the limit - then, the block will spawn
 * another one:
 * x-xxXX-xx-x-xxxxxxx-x
 *
 * Again, we don't touch other blocks during deletion. Only when the block
 * size hits zero, we kill it:
 * x-xxXX-xx-x-xxxxxx-x
 *
 * Note that we still keep around blocks which have any counterpart in the
 * file, even if they are already dead in our current file image. This is
 * necessary when reading more file blocks to the cache and when commiting
 * the changes, in order to maintain offset consistency.
 *
 * When can size != phys_size?
 * * byte(s) have been INSERTED into this block: size > phys_size
 *	Special case is an entire new block:
 *		in case of virtual blocks, phys_size MUST == 0 (gap)
 * * byte(s) have been ERASED from this block: size < phys_size
 *	Virtual blocks are not permitted
 */

struct file_block {
#ifdef USE_TREE
	struct tree_head t;
#else
	LIST_HEAD(struct file_block);
	struct {
		/* Size of the @src array (or the represented gap, for virtual
		 * blocks) - as the file is currently in the memory. */
		off_t size;
		/* Corresponding size in the file as currently laid out on the disk. */
		off_t phys_size;
	} t;
#endif

	/* Whether the block is modified in the memory. */
	bool dirty;
	/* Whether the block is in NOT the cache[] array
	 * (if so, it can be free()d after it is undirtied). */
	bool excache;
	/* "Virtual" block - it must never be dirty nor excache
	 * and src is not valid - only size is meaningful. These
	 * blocks are always dynamically allocated, NEVER in the
	 * cache[] array. */
	/* TODO: virtual and size should be at the start and the rest
	 * shouldn't be allocated then. */
	bool virtual;

	/* Poor man's refcount; if non-zero, the block will not
	 * be auto-removed by the utility routines. We do not do
	 * full refcounting, only quickly bump this when calling
	 * something dangerous and then set it back to zero. */
	int protect;

	unsigned char src[FILE_BLOCK_SIZE];
};

struct file {
	const unsigned char *name;
	struct stat s;
	off_t size;
	int fd;
	bool modified;

	/* As the blocks are read, they are cycling the ring cache. */
	int cache_pos;
	struct file_block cache[CACHE_LENGTH];
	/* The file blocks are sorted by the offset in this list. */
#ifdef USE_TREE
	struct tree blocks;
#else
	struct list_head blocks;
#endif

	bool marks_set['z' - 'a' + 1 + 'Z' - 'A' + 1];
	off_t marks['z' - 'a' + 1 + 'Z' - 'A' + 1];
};

static inline char
ch2mark(char ch)
{
	return ('a' <= ch && ch <= 'z') ? ch - 'a' :
	       ('A' <= ch && ch <= 'Z') ? 26 + ch - 'A' :
	       -1;
}

#ifdef USE_TREE

#define first_block(f) ((struct file_block *) first_in_tree(&(f)->blocks))
#define prev_block(f,b) ((struct file_block *) prev_in_tree(&(f)->blocks, (b)))
#define next_block(f,b) ((struct file_block *) next_in_tree(&(f)->blocks, (b)))
#define last_block(f) ((struct file_block *) last_in_tree(&(f)->blocks))

/* false if this is not a block - out of list bounds */
#define is_a_block(f,b) (!!(b))
/* Whether the block is a blocks list member. */
#define is_chained_block(f,b) ((b)->t.up || (f)->blocks.root == (b))

#define chain_block(f,b) do { add_to_tree(&(f)->blocks, 0, (b)); } while (0)
#define recalc_chain_block(f,b) do { \
	chain_block((f), (b)); \
	recalc_node_recursive((b)); \
} while (0)
#define chain_block_after(f,p,b) do { add_after_in_tree(&(f)->blocks, (p), (b)); } while (0)
#define recalc_chain_block_after(f,p,b) do { \
	chain_block_after((f), (p), (b)); \
	recalc_node_recursive((b)); \
} while (0)
#define unchain_block(f,b) do { unchain_from_tree(&(f)->blocks, (b)); } while (0)
#define recalc_unchain_block(f,b) do { del_from_tree(&(f)->blocks, (b)); } while (0)
#define init_block_list(f) do { f->blocks = *init_tree(); } while (0)

#define foreach_block(b,f) foreach_in_tree (b, &(f)->blocks)
#define foreachsafe_block(b,n,f) foreachsafe_in_tree (b, n, &(f)->blocks)

#else

#define first_block(f) ((struct file_block *) (f)->blocks.next)
#define prev_block(f,b) (b)->prev
#define next_block(f,b) (b)->next
#define last_block(f) ((struct file_block *) (f)->blocks.prev)

/* false if this is not a block - out of list bounds */
#define is_a_block(f,b) ((b) != (struct file_block *) &(f)->blocks)
/* Whether the block is a blocks list member. */
#define is_chained_block(f,b) !!((b)->next)

#define recalc_node_recursive(b)
#define chain_block(f,b) do { add_to_list((f)->blocks, (b)); } while (0)
#define chain_block_after(f,p,b) do { add_at_pos((p), (b)); } while (0)
#define recalc_chain_block_after(f,p,b) do { chain_block_after((f), (p), (b)); } while (0)
#define unchain_block(f,b) do { del_from_list(b); (b)->next = NULL; } while (0)
#define init_block_list(f) do { init_list((f)->blocks); } while (0)

#define foreach_block(b,f) foreach (b, (f)->blocks)
#define foreachsafe_block(b,n,f) foreachsafe (b, n, (f)->blocks)

#endif


/* Opens a file. */
struct file *file_init(const unsigned char *name);

static off_t file_size(struct file *file);

/* Fetches the required part of the file. The real size is saved to *size.
 * On error, the *size will be < 0 and the call will return NULL. The error
 * will be available in @errno at that time.
 *
 * Please note that even for incomplete reads, the memory pointed at
 * by the return value will be allocated to the full size as specified
 * by *size. So shall the buffer be held for a longer time, you might
 * want to reallocate the buffer to the exact length. */
void *file_fetch_block(struct file *file, off_t offset, ssize_t *size);

/* Find a given sequence of bytes in the file, starting at the @from file
 * offset and possibly wrapping around. @dir is +1 or -1. */
off_t file_find_expr(struct file *file, off_t from, struct expression *expr, reg_callback expr_cb, void *expr_cb_data, int dir);
#define FINDOFF_NO_MATCH (-1)

/* Sets a single byte in the file. */
void file_set_byte(struct file *file, off_t offset, unsigned char byte);

/* Inserts a single byte to the file. */
void file_insert_byte(struct file *file, off_t offset, unsigned char byte);
/* Insert a block of bytes to the file. */
void file_insert_block(struct file *file, off_t offset, unsigned char *buf, off_t len);

/* Erases a single byte from the file. */
void file_erase_byte(struct file *file, off_t offset);
/* Erases a block of bytes from the file. */
void file_erase_block(struct file *file, off_t offset, off_t len);

/* Commits dirty blocks to the file. */
void file_commit(struct file *file);

/* Closes the file. */
void file_done(struct file *file);

/* Saves dump of internal structures to disk. */
void file_swp(struct file *file);

/* Loads dump of internal structures associated with given file from disk. */
struct file *file_unswp(const unsigned char *name);

/* Remoes dump of internal structures associated with given file from disk. */
void file_rmswp(const unsigned char *name);


static inline off_t
file_size(struct file *file)
{
	return file->size;
}


#endif
