/* $Id$ */

/*
 * hed - Hexadecimal editor
 * Copyright (C) 2004  Petr Baudis <pasky@ucw.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * When his eyes were in turn uncovered, Frodo looked up and caught his breath.
 * They were standing in an open space. To the left stood a great mound,
 * covered with a sward of grass as green as Spring-time in the Elder Days.
 * Upon it, as a double crown, grew two circles of trees: the outer had bark of
 * snowy white, and were leafless but beautiful in their shapely nakedness; the
 * inner were mallorn-trees of great height, still arrayed in pale gold. High
 * amid the branches of a towering tree that stood in the centre of all there
 * gleamed a white flet. At the feet of the trees, and all about the green
 * hillsides the grass was studded with small golden flowers shaped like stars.
 * Among them, nodding on slender stalks, were other flowers, white and palest
 * green: they glimmered as a mist amid the rich hue of the grass. Over all the
 * sky was blue, and the sun of afternoon glowed upon the hill and cast long
 * green shadows beneath the trees.
 */

#include "config.h"

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>

#include "hed.h"

#include "util/tree.h"


//#define TREE_DEBUG
#ifdef TREE_DEBUG
#define TDEBUG(x...) fprintf(stderr, x)
#else
#define TDEBUG(x...)
#endif


struct tree *
init_tree(void)
{
	struct tree *tree = calloc(1, sizeof(struct tree));
	tree->root = NULL;
	return tree;
}

void
recalc_node(void *x)
{
	struct tree_head *item = x;

	item->cover_size = item->size;
	item->cover_phys_size = item->phys_size;
	if (item->left) {
		item->cover_size += item->left->cover_size;
		item->cover_phys_size += item->left->cover_phys_size;
		item->left->up = item;
	}
	if (item->right) {
		item->cover_size += item->right->cover_size;
		item->cover_phys_size += item->right->cover_phys_size;
		item->right->up = item;
	}
}

void
recalc_node_recursive(void *x)
{
	struct tree_head *item = x;
	while (item) {
		recalc_node(item);
		item = item->up;
	}
}

struct tree_head *
unchain_from_tree(struct tree *tree, void *x)
{
	struct tree_head *item = x;
	struct tree_head *parent;

	if (!item->left && !item->right && !item->up) {
		tree->root = NULL;
		return NULL;
	}

	unsplay(tree, item);

	assert(!item->left && !item->right && item->up);
	if (item->up->left == item) {
		item->up->left = NULL;
	} else {
		assert(item->up->right == item);
		item->up->right = NULL;
	}

	parent = item->up;
	item->up = NULL;
	return parent;
}

void
del_from_tree(struct tree *tree, void *x)
{
	struct tree_head *item = x;
	struct tree_head *parent;

	/*
	 * 'I'll give your name and number to the Nazgyl,' said the soldier
	 * lowering his voice to a hiss. 'One of *them*'s in charge at the
	 * Tower now.'
	 */

	parent = unchain_from_tree(tree, x);
	assert(item->size == item->cover_size);
	while (parent) {
		parent->cover_size -= item->size;
		parent->cover_phys_size -= item->phys_size;
		parent = parent->up;
	}
}

static void
chain_in_tree(struct tree *tree, struct tree_head *pos, struct tree_head *item, bool splayful)
{
	item->up = item->left = item->right = NULL;
	item->cover_size = item->size;
	item->cover_phys_size = item->phys_size;

	if (splayful)
		splay(tree, pos ? pos : first_in_tree(tree));
	if (!pos) {
		pos = tree->root;
		item->left = NULL;
		item->right = pos;
	} else {
		assert(tree->root == pos);
		item->left = pos;
		item->right = pos->right;
		pos->right = NULL;
		recalc_node(pos);
	}
	if (pos)
		pos->up = item;
	tree->root = item;
	recalc_node(item);
}

void
add_after_in_tree(struct tree *tree, void *y, void *x)
{
	return chain_in_tree(tree, y, x, true);
}

void
add_to_tree(struct tree *tree, off_t offset, void *x)
{
	return chain_in_tree(tree, find_in_tree(tree, offset), x, false);
}

void *
find_in_tree(struct tree *tree, off_t offset)
{
	struct tree_head *item = NULL, *nitem = tree->root;
	off_t pos = nitem && nitem->left ? nitem->left->cover_size : 0;

	while (nitem) {
		/* We splay as we go. */
		TDEBUG("find_in_tree(%llx): %llx\n", offset, pos);
		/* rot_up(tree, nitem); */
		item = nitem;
		if (pos <= offset && offset < pos + item->size) {
			TDEBUG("find_in_tree(%llx): match - pos %llx, size %llx\n", offset, pos, item->size);
			nitem = NULL;
		} else if (pos > offset) {
			TDEBUG("find_in_tree(%llx): going left\n", offset);
			nitem = item->left;
			pos -= nitem ? nitem->size + (nitem->right ? nitem->right->cover_size : 0) : 0;
		} else { assert(pos < offset && pos + item->size <= offset);
			TDEBUG("find_in_tree(%llx): going right\n", offset);
			nitem = item->right;
			pos += item->size + (nitem && nitem->left ? nitem->left->cover_size : 0);
			/* If we are going far too left, we must return NULL. */
			item = nitem;
		}
	}

	if (item)
		TDEBUG("find_in_tree(%llx): chosen %llx + %llx\n", offset, pos, item->size);
	else
		TDEBUG("find_in_tree(%llx): chosen %llx - but not found\n", offset, pos);

	/* TODO: Splay as we go? */
	if (item)
		splay(tree, item);

	return item;
}

void *
first_in_tree(struct tree *tree)
{
	struct tree_head *item = tree->root;

	if (!item) return NULL;
	while (item->left)
		item = item->left;
	return item;
}

/* XXX: It would be more logical to prev(NULL) = last, next(NULL) = first.
 * Why don't we do it? */

void *
prev_in_tree(struct tree *tree, void *x)
{
	struct tree_head *item = x;

	if (!item) return first_in_tree(tree);
	splay(tree, item);
	item = item->left;
	if (!item) return NULL;
	while (item->right)
		item = item->right;
	return item;
}

void *
next_in_tree(struct tree *tree, void *x)
{
	struct tree_head *item = x;

	if (!item) return NULL; /* TODO: last_in_tree()? --pasky */
	splay(tree, item);
	item = item->right;
	if (!item) return NULL;
	while (item->left)
		item = item->left;
	return item;
}

void *
last_in_tree(struct tree *tree)
{
	struct tree_head *item = tree->root;

	if (!item) return NULL;
	while (item->right)
		item = item->right;
	return item;
}

off_t
tree_block_offset(void *x)
{
	struct tree_head *item = x;
	off_t off = 0;

	assert(item);
	do {
		if (item->left)
			off += item->left->cover_size;
		item = item->up;
	} while (item);
	return off;
}

off_t
tree_block_phys_offset(void *x)
{
	struct tree_head *item = x;
	off_t off = 0;

	assert(item);
	do {
		if (item->left)
			off += item->left->cover_phys_size;
		item = item->up;
	} while (item);
	return off;
}


static void
update_parent(struct tree *tree, struct tree_head *old, struct tree_head *new)
{
	new->up = old->up;
	if (!new->up) {
		assert(tree->root == old);
		tree->root = new;
	} else {
		if (old->up->left == old) {
			new->up->left = new;
		} else { assert(old->up->right == old);
			new->up->right = new;
		}
	}
}

void
splay(struct tree *tree, void *x)
{
	struct tree_head *item = x;
	struct tree_head *item_mid, *item_top;

	assert(item);
	assert(tree->root);
	if (tree->root == item)
		return;

	assert(item->up);
	if (item->up == tree->root) {
		if (item->up->left == item) {
			item->up->left = item->right;
			item->right = item->up;
			recalc_node(item->up);
			recalc_node(item);
		} else { assert(item->up->right == item);
			item->up->right = item->left;
			item->left = item->up;
			recalc_node(item->up);
			recalc_node(item);
		}
		tree->root = item->up->up = item;
		item->up = NULL;
		return;
	}

	/*
	 * 'Come on, you slugs!' he cried. 'This is no time for slouching.' He
	 * took a step towards them, and even in the gloom he recognized the
	 * devices on their shields. 'Deserting, eh?' he snarled. 'Or thinking
	 * of it? All your folk should have been inside Udyn before yesterday
	 * evening. You know that. Up you get and fall in, or I'll have your
	 * numbers and report you.'
	 */

	/* TODO: Something better than recursion. */
	item_mid = item->up; assert(item_mid);
	item_top = item_mid->up; assert(item_top);

	if (item_top->left == item_mid && item_mid->left == item) {
		update_parent(tree, item_top, item);
		item_mid->left = item->right;
		item->right = item_top;
		item_top->up = item;
		goto splay_on;
	}
	/* Inverse operation. */
	if (item_top->right == item_mid && item_mid->right == item) {
		update_parent(tree, item_top, item);
		item_mid->right = item->left;
		item->left = item_top;
		item_top->up = item;
		goto splay_on;
	}
	
	if (item_top->left == item_mid && item_mid->right == item) {
		update_parent(tree, item_top, item);
		item_top->up = item;
		item_top->left = item->right;
		item->right = item_top;

		item_mid->up = item;
		item_mid->right = item->left;
		item->left = item_mid;
		goto splay_on;
	}
	/* Inverse operation. */
	if (item->up->left == item && item->up->up->right == item->up) {
		update_parent(tree, item_top, item);
		item_top->up = item;
		item_top->right = item->left;
		item->left = item_top;

		item_mid->up = item;
		item_mid->left = item->right;
		item->right = item_mid;
		goto splay_on;
	}
	assert(0);
	return;

splay_on:
	recalc_node(item_mid);
	recalc_node(item_top);
	recalc_node(item);
	splay(tree, item);
}

void
unsplay(struct tree *tree, void *x)
{
	struct tree_head *item = x;
	struct tree_head *item_down;
	enum { LEAF_LEFT, LEAF_RIGHT } leaf;

	if (!tree->root || (!item->left && !item->right))
		return;

	/* TODO: If we kept track of how many items are in each
	 * subtree, we could've made an actually inteligent
	 * choice. */
	/* TODO: We could at least randomize it? */

	if (item->left) {
		item_down = item->left;
		leaf = LEAF_LEFT;
	} else { assert(item->right);
		item_down = item->right;
		leaf = LEAF_RIGHT;
	}

	assert(item_down->up == item);
	update_parent(tree, item, item_down);
	item->up = item_down;
	if (leaf == LEAF_LEFT) {
		item->left = item_down->right;
		item_down->right = item;
	} else { assert(leaf == LEAF_RIGHT);
		item->right = item_down->left;
		item_down->left = item;
	}
	recalc_node(item);
	recalc_node(item_down);
	unsplay(tree, item);
}
